from typing import Optional

from qm.pb.inc_qua_pb2 import QuaProgram
from qm.program.StatementsCollection import StatementsCollection
from qm.program._ResultAnalysis import _ResultAnalysis


class _Program:
    def __init__(self, config=None, program: Optional[QuaProgram] = None):
        super().__init__()
        self._program = program if program is not None else QuaProgram()
        self._program.script.SetInParent()
        self._program.resultAnalysis.SetInParent()
        self._program.script.body.SetInParent()
        self._qua_config = config
        self._result_analysis = _ResultAnalysis(self._program.resultAnalysis)
        self._is_in_scope = False

    def _declare_var(self, name, var_type, size, value, dim, is_input_stream):
        declaration = self._program.script.variables.add()
        declaration.name = name
        declaration.type = var_type
        declaration.size = size
        declaration.dim = dim
        declaration.isInputStream = is_input_stream
        if value is None:
            pass
        elif type(value) is list:
            for i in value:
                added_value = declaration.value.add()
                added_value.CopyFrom(i)
        else:
            added_value = declaration.value.add()
            added_value.CopyFrom(value)

    def declare_int(self, name, size, value, dim, is_input_stream):
        self._declare_var(name, QuaProgram.INT, size, value, dim, is_input_stream)

    def declare_real(self, name, size, value, dim, is_input_stream):
        self._declare_var(name, QuaProgram.REAL, size, value, dim, is_input_stream)

    def declare_bool(self, name, size, value, dim, is_input_stream):
        self._declare_var(name, QuaProgram.BOOL, size, value, dim, is_input_stream)

    @property
    def body(self):
        return StatementsCollection(self._program.script.body)

    @property
    def result_analysis(self) -> _ResultAnalysis:
        return self._result_analysis

    def build(self, config):
        copy = QuaProgram()
        copy.CopyFrom(self._program)
        copy.config.CopyFrom(config)
        return copy

    def set_in_scope(self):
        self._is_in_scope = True

    def set_exit_scope(self):
        self._is_in_scope = False

    def is_in_scope(self) -> bool:
        return self._is_in_scope



import functools
import warnings


class QMDeprecationWarning(Warning):
    pass


def deprecated(reason: str, version: str, last: str):
    """
    Used to mark API as deprecated
    The reason of deprecation should be specified:

    - if there is an alternative, this is a reason enough
    - otherwise, the reason can be empty

    It holds the version it was deprecated on, and the version it will no longer be available

    Automatic testing should verify that it doesn't exist on version that it shouldn't exist
    """

    # TODO: throw exception if current version > last

    def deprecated_decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            warnings.warn(
                f"{func.__qualname__} is deprecated since {version}: {reason}",
                category=QMDeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return func_wrapper

    return deprecated_decorator



from enum import Enum
from typing import Optional, List

from google.protobuf import empty_pb2

from qm.QmPendingJob import QmPendingJob
from qm._errors import FailedToAddJobToQueueException
from qm.pb.frontend_pb2 import (
    JobQueryParams,
    QueryValueMatcher,
    AddToQueueRequest,
    QueuePosition,
    AddCompiledToQueueRequest,
)
import qm.pb.frontend_pb2
from qm.pb.inc_qua_pb2 import QuaProgram
from qm.logger import logger
from qm.program import _Program
from qm.program._execution_overrides_schema import ExecutionOverridesSchema
from qm.utils import _level_map, _set_compiler_options


class JobNotFoundError(Exception):
    pass


class _QueuePosition(Enum):
    Start = 1
    End = 2


class QmQueue(object):
    def __init__(
        self, machine, qmm: "qm.QuantumMachinesManager.QuantumMachinesManager"
    ):
        self._machine = machine
        self._qmm = qmm

    def _get_pending_jobs(
        self, job_id: str = None, position: int = None, user_id: str = None
    ) -> List[QmPendingJob]:
        request = JobQueryParams()
        request.quantumMachineId = self._machine.id
        if job_id is not None:
            job_query_value_matcher = QueryValueMatcher()
            job_query_value_matcher.value = job_id
            request.jobId.CopyFrom(job_query_value_matcher)
        if position is not None:
            request.position.value = position
        if user_id is not None:
            user_query_value_matcher = QueryValueMatcher()
            user_query_value_matcher.value = user_id
            request.userId.CopyFrom(user_query_value_matcher)

        response = self._qmm._frontend.GetPendingJobs(request)
        jobs = [
            (i, response.pendingJobs[i].positionInQueue)
            for i in list(response.pendingJobs)
        ]
        jobs.sort(key=lambda it: it[1])
        result = [
            QmPendingJob(pending_job_id, self._machine, self._qmm)
            for pending_job_id, pos in jobs
        ]
        return result

    def add(self, program: _Program, **kwargs) -> QmPendingJob:
        """
        Adds a QmJob to the queue.
        Programs in the queue will play as soon as possible.

        :param program: A QUA program
        :return: The QmJob

        Example::

        >>> qm.queue.add(program)  # adds at the end of the queue
        >>> qm.queue.insert(program, position)  # adds at position
        """
        return self._insert(program, _QueuePosition.End, **kwargs)

    def add_compiled(self, program_id: str, overrides=None) -> QmPendingJob:
        """
        Adds a compiled QUA program to the end of the queue, optionally
        overriding the values of analog waveforms defined in the program.
        Programs in the queue will play as soon as possible.
        For a detailed explanation see
        `Precompile Jobs <https://qm-docs.qualang.io/guides/features.html#precompile-jobs>`__.

        :param program_id: A QUA program ID returned from the compile function
        :param overrides: Object containing Waveforms to run the program with

        Example::

        >>> program_id = qm.compile(...)
        >>> pending_job = qm.queue.add_compiled(program_id, overrides={
        >>>     'waveforms': {
        >>>         'my_arbitrary_waveform': [0.1, 0.2, 0.3],
        >>>         'my_constant_waveform': 0.2
        >>>     }
        >>> })
        >>> job = pending_job.wait_for_execution()
        """
        if not overrides:
            overrides = {}

        return self._insert_compiled(program_id, overrides)

    def add_to_start(self, program: _Program, **kwargs) -> QmPendingJob:
        """
        Adds a QMJob to the start of the queue.
        Programs in the queue will play as soon as possible.

        :param program: A QUA program
        :return: The QmJob

        """
        return self._insert(program, _QueuePosition.Start, **kwargs)

    def _insert(
        self, program: _Program, position: _QueuePosition, **kwargs
    ) -> QmPendingJob:
        return self._insert_pb(
            program.build(self._machine._pb_config), position, **kwargs
        )

    def _parse_add_compiled_to_queue_response(self, response) -> QmPendingJob:
        job_id = response.jobId

        for err in response.errors:
            logger.error(err.message)

        if not response.ok:
            logger.error("Job " + job_id + " failed. Failed to execute program.")
            raise FailedToAddJobToQueueException(job_id)

        return QmPendingJob(job_id, self._machine, self._qmm)

    def _parse_add_to_queue_response(self, response) -> QmPendingJob:
        messages = [(_level_map[msg.level], msg.message) for msg in response.messages]

        job_id = response.jobId

        for lvl, msg in messages:
            logger.log(lvl, msg)

        if not response.ok:
            logger.error("Job " + job_id + " failed. Failed to execute program.")
            raise FailedToAddJobToQueueException(job_id)

        return QmPendingJob(job_id, self._machine, self._qmm)

    def _insert_compiled(self, program_id, overrides):
        request_queue_position = QueuePosition()
        request_queue_position.end.CopyFrom(empty_pb2.Empty())

        request = AddCompiledToQueueRequest(
            queuePosition=request_queue_position,
            quantumMachineId=self._machine.id,
            programId=program_id,
            executionOverrides=ExecutionOverridesSchema().load(overrides),
        )

        logger.info("Sending pre-compiled program to QOP")

        response = self._qmm._frontend.AddCompiledToQueue(request)
        return self._parse_add_compiled_to_queue_response(response)

    def _insert_pb(
        self, program: QuaProgram, position: _QueuePosition, **kwargs
    ) -> QmPendingJob:
        request = AddToQueueRequest()
        request.quantumMachineId = self._machine.id
        request.highLevelProgram.CopyFrom(program)

        _set_compiler_options(request, **kwargs)

        queue_position = QueuePosition()
        if position == _QueuePosition.Start:
            queue_position.start.CopyFrom(empty_pb2.Empty())
        elif position == _QueuePosition.End:
            queue_position.end.CopyFrom(empty_pb2.Empty())

        logger.info("Sending program to QOP")

        request.queuePosition.CopyFrom(queue_position)
        response = self._qmm._frontend.AddToQueue(request)
        return self._parse_add_to_queue_response(response)

    @property
    def count(self):
        """
        Get the number of jobs currently on the queue

        :return: The number of jobs in the queue

        Example::

        >>> qm.queue.count
        """
        return len(self._get_pending_jobs())

    def __len__(self):
        return self.count

    @property
    def pending_jobs(self) -> List[QmPendingJob]:
        """
        Returns all currently pending jobs

        :return: A list of all of the currently pending jobs
        """
        return self._get_pending_jobs()

    def get(self, job_id) -> QmPendingJob:
        """
        Get a pending job object by job_id

        :param job_id: a QMJob id
        :return: The pending job

        Example::

        >>> qm.queue.get(job_id)

        """
        jobs = self._get_pending_jobs(job_id)
        if len(jobs) == 0:
            raise JobNotFoundError()
        return jobs[0]

    def get_at(self, position: int) -> QmPendingJob:
        """
        Gets the pending job object at the given position in the queue

        :param position: An integer position in queue
        :return: The pending job

        Example::

        >>> qm.queue.get(job_id)

        """
        jobs = self._get_pending_jobs(None, position)
        if len(jobs) == 0:
            raise JobNotFoundError()
        return jobs[0]

    def get_by_user_id(self, user_id: str) -> List[QmPendingJob]:
        return self._get_pending_jobs(None, None, user_id)

    def remove_by_id(self, job_id: str) -> int:
        """
        Removes the pending job object with a specific job id

        :param job_id: a QMJob id
        :return: The number of jobs removed

        Example::

        >>> qm.queue.remove_by_id(job_id)
        """
        if job_id is None or job_id == "":
            raise ValueError("job_id can not be empty")
        return self._remove(job_id=job_id, position=None, user_id=None)

    def remove_by_position(self, position: int) -> int:
        """
        Remove the PendingQmJob object by position in queue

        :param position: position in queue
        :return: The number of jobs removed

        Example::

        >>> qm.queue.remove_by_position(position)

        """
        if position is None or position <= 0:
            raise ValueError("position must be positive")
        return self._remove(job_id=None, position=position, user_id=None)

    def remove_by_user_id(self, user_id: str):
        return self._remove(job_id=None, position=None, user_id=user_id)

    def _remove(
        self,
        job_id: Optional[str] = None,
        position: Optional[int] = None,
        user_id: Optional[str] = None,
    ) -> int:
        request = JobQueryParams()
        request.quantumMachineId = self._machine.id
        if job_id is not None:
            job_query_value_matcher = QueryValueMatcher()
            job_query_value_matcher.value = job_id
            request.jobId.CopyFrom(job_query_value_matcher)

        if position is not None:
            request.position.value = position

        if user_id is not None:
            user_query_value_matcher = QueryValueMatcher()
            user_query_value_matcher.value = user_id
            request.userId.CopyFrom(user_query_value_matcher)

        response = self._qmm._frontend.RemovePendingJobs(request)
        return response.numbersOfJobsRemoved

    def __getitem__(self, position) -> QmPendingJob:
        return self.get_at(position)

    def clear(self) -> int:
        """
        Empties the queue from all pending jobs

        :return: The number of jobs removed
        """
        return self._remove()



from qm.pb.inc_qua_pb2 import QuaResultAnalysis

_ResultSymbol = "@re"


class _ResultAnalysis:
    def __init__(self, result_analysis: QuaResultAnalysis):
        super().__init__()
        self._result_analysis = result_analysis
        self._saves = []

    def get_outputs(self):
        names = []
        for save in self._saves:
            if save.tag is not None:
                names.append(save.tag)
        return names

    def save(self, tag, expression):
        for save in self._saves:
            if save.tag == tag:
                raise Exception("can not save two streams with the same tag")
        self._saves.append(_OutputStream(expression, ["save", tag], tag))

    def save_all(self, tag, expression):
        for save in self._saves:
            if save.tag == tag:
                raise Exception("can not save two streams with the same tag")
        self._saves.append(_OutputStream(expression, ["saveAll", tag], tag))

    def auto_save_all(self, tag, expression):
        for save in self._saves:
            if save.tag == tag:
                raise Exception("can not save two streams with the same tag")
        self._saves.append(_OutputStream(expression, ["saveAll", tag, "auto"], tag))

    def _add_pipeline(self, output):
        pipeline = self._result_analysis.model.add()
        pipeline.extend(output._to_proto())

    def generate_proto(self):
        for save in self._saves:
            self._add_pipeline(save)
        # print(self._result_analysis.model)

    def build(self):
        copy = QuaResultAnalysis()
        copy.CopyFrom(self._result_analysis)
        return copy


class _OutputStream(object):
    def __init__(self, input_stream, operator_array, tag):
        super().__init__()
        self._input_stream = input_stream
        self._operator_array = operator_array
        self.tag = tag

    def _to_proto(self):
        return self._operator_array + [self._input_stream._to_proto()]

    def input_stream(self):
        return self._input_stream

    def set_input_stream(self, new_stream):
        self._input_stream = new_stream


class _Node(object):
    def __init__(self, input_stream, output):
        self._input_stream = input_stream
        self._outputs = [output]

    def input_stream(self):
        return self._input_stream

    def outputs(self):
        return self._outputs



import grpc

# taken from https://github.com/grpc/grpc/tree/master/examples/python/auth

_SIGNATURE_HEADER_KEY = "x-signature"


class AuthGateway(grpc.AuthMetadataPlugin):
    def __call__(self, context, callback):
        """Implements authentication by passing metadata to a callback.
        Implementations of this method must not block.
        Args:
          context: An AuthMetadataContext providing information on the RPC that
            the plugin is being called to authenticate.
          callback: An AuthMetadataPluginCallback to be invoked either
            synchronously or asynchronously.
        """
        # Example AuthMetadataContext object:
        # AuthMetadataContext(
        #     service_url=u'https://localhost:50051/helloworld.Greeter',
        #     method_name=u'SayHello')
        signature = context.method_name[::-1]
        callback(((_SIGNATURE_HEADER_KEY, signature),), None)


def create_credentials():
    import certifi

    # Call credential object will be invoked for every single RPC
    call_credentials = grpc.metadata_call_credentials(
        AuthGateway(), name="auth gateway"
    )
    # Channel credential will be valid for the entire channel
    ROOT_CERTIFICATE = _load_credential_from_file(certifi.where())
    channel_credential = grpc.ssl_channel_credentials(ROOT_CERTIFICATE)
    # Combining channel credentials and call credentials together
    composite_credentials = grpc.composite_channel_credentials(
        channel_credential,
        call_credentials,
    )
    return composite_credentials


def _load_credential_from_file(filepath):
    with open(filepath, "rb") as f:
        return f.read()


__all__ = ["create_credentials"]



from typing import Tuple

import qm.pb.errors_pb2


class QmApiError(BaseException):
    def __init__(self, code: int, message: str = "") -> None:
        super().__init__(message)
        self.message = message
        self.code = code


class UnspecifiedError(QmApiError):
    def __init__(self, message: str) -> None:
        super().__init__(0, message)


class _QmJobError(QmApiError):
    """Base class for exceptions in this module."""

    pass


class MissingJobError(_QmJobError):
    """
    if the job isn't recognized by the server (can happen if it never ran or if it was already deleted)
    """

    def __init__(self) -> None:
        super().__init__(1000)


class InvalidJobExecutionStatusError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1001)


class InvalidOperationOnSimulatorJobError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1002)


class InvalidOperationOnRealJobError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1003)


class UnknownInputStreamError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1006)


class ConfigQueryError(QmApiError):
    pass


class MissingElementError(ConfigQueryError):
    """ """

    @staticmethod
    def _build_from_response(request, response) -> "MissingElementError":
        return MissingElementError(
            response.jobManagerResponseHeader.jobErrorDetails.message
        )

    def __init__(self, message: str):
        super().__init__(4001, message)


class MissingDigitalInputError(ConfigQueryError):
    """"""

    @staticmethod
    def _build_from_response(request, response) -> "MissingDigitalInputError":
        return MissingDigitalInputError(
            response.jobManagerResponseHeader.jobErrorDetails.message
        )

    def __init__(self, message: str):
        super().__init__(4002, message)


class _InvalidConfigChangeError(QmApiError):
    pass


class ElementWithSingleInputError(_InvalidConfigChangeError):
    @staticmethod
    def _build_from_response(request, response) -> "ElementWithSingleInputError":
        return ElementWithSingleInputError(request.qeName)

    def __init__(self, element_name: str):
        super().__init__(3000)
        self.element_name = element_name


class InvalidElementCorrectionError(_InvalidConfigChangeError):
    """
    If the correction values are invalid
    """

    @staticmethod
    def _build_from_response(request, response) -> "InvalidElementCorrectionError":
        return InvalidElementCorrectionError(
            response.jobManagerResponseHeader.jobErrorDetails.message,
            request.qeName,
            (
                request.correction.v00,
                request.correction.v01,
                request.correction.v10,
                request.correction.v11,
            ),
        )

    def __init__(
        self,
        message: str,
        element_name: str,
        correction: Tuple[float, float, float, float],
    ) -> None:
        super().__init__(3001, message)
        self.element_name = element_name
        self.correction = correction


class ElementWithoutIntermediateFrequencyError(_InvalidConfigChangeError):
    @staticmethod
    def _build_from_response(
        request, response
    ) -> "ElementWithoutIntermediateFrequencyError":
        return ElementWithoutIntermediateFrequencyError(request.qeName)

    def __init__(self, element_name: str):
        super().__init__(3002)
        self.element_name = element_name


class InvalidDigitalInputThresholdError(_InvalidConfigChangeError):
    @staticmethod
    def _build_from_response(request, response) -> "InvalidDigitalInputThresholdError":
        return InvalidDigitalInputThresholdError(
            response.jobManagerResponseHeader.jobErrorDetails.message
        )

    def __init__(self, message: str):
        super().__init__(3003)
        self.message = message


class InvalidDigitalInputWindowError(_InvalidConfigChangeError):
    @staticmethod
    def _build_from_response(request, response) -> "InvalidDigitalInputWindowError":
        return InvalidDigitalInputWindowError(
            response.jobManagerResponseHeader.jobErrorDetails.message
        )

    def __init__(self, message: str):
        super().__init__(3004)
        self.message = message


class InvalidDigitalInputPolarityError(_InvalidConfigChangeError):
    @staticmethod
    def _build_from_response(request, response) -> "InvalidDigitalInputPolarityError":
        return InvalidDigitalInputPolarityError(
            response.jobManagerResponseHeader.jobErrorDetails.message
        )

    def __init__(self, message: str):
        super().__init__(3005)
        self.message = message


def _handle_job_manager_error(request, response, valid_errors):
    api_response = response.jobManagerResponseHeader
    if not api_response.success:
        error_type = api_response.jobManagerErrorType

        if error_type == qm.pb.errors_pb2.MissingJobError:
            raise MissingJobError()
        elif error_type == qm.pb.errors_pb2.InvalidJobExecutionStatusError:
            raise InvalidJobExecutionStatusError()
        elif error_type == qm.pb.errors_pb2.InvalidOperationOnSimulatorJobError:
            raise InvalidOperationOnSimulatorJobError()
        elif error_type == qm.pb.errors_pb2.InvalidOperationOnRealJobError:
            raise InvalidOperationOnRealJobError()
        elif error_type == qm.pb.errors_pb2.JobOperationSpecificError:
            exception_to_raise = _get_handle_job_operation_error(request, response)
            if (
                exception_to_raise is not None
                and type(exception_to_raise) in valid_errors
            ):
                raise exception_to_raise
            else:
                raise UnspecifiedError("Unspecified operation specific error")
        elif error_type == qm.pb.errors_pb2.ConfigQueryError:
            exception_to_raise = _get_handle_config_query_error(request, response)
            if (
                exception_to_raise is not None
                and type(exception_to_raise) in valid_errors
            ):
                raise exception_to_raise
            else:
                raise UnspecifiedError("Unspecified operation specific error")
        elif error_type == qm.pb.errors_pb2.UnknownInputStreamError:
            raise UnknownInputStreamError()
        else:
            raise UnspecifiedError("Unspecified operation error")


def _get_handle_config_query_error(request, response):
    error_type = response.jobManagerResponseHeader.jobErrorDetails.configQueryErrorType
    exception_to_raise = None
    if error_type == qm.pb.errors_pb2.MissingElementError:
        exception_to_raise = MissingElementError._build_from_response(request, response)
    elif error_type == qm.pb.errors_pb2.MissingDigitalInputError:
        exception_to_raise = MissingDigitalInputError._build_from_response(
            request, response
        )
    else:
        exception_to_raise = UnspecifiedError("Unspecified config query error")

    return exception_to_raise


def _get_handle_job_operation_error(request, response):
    error_type = (
        response.jobManagerResponseHeader.jobErrorDetails.jobOperationSpecificErrorType
    )

    if error_type == qm.pb.errors_pb2.SingleInputElementError:
        exception_to_raise = ElementWithSingleInputError._build_from_response(
            request, response
        )
    elif error_type == qm.pb.errors_pb2.InvalidCorrectionMatrixError:
        exception_to_raise = InvalidElementCorrectionError._build_from_response(
            request, response
        )
    elif error_type == qm.pb.errors_pb2.ElementWithoutIntermediateFrequencyError:
        exception_to_raise = (
            ElementWithoutIntermediateFrequencyError._build_from_response(
                request, response
            )
        )
    elif error_type == qm.pb.errors_pb2.InvalidDigitalInputThresholdError:
        exception_to_raise = InvalidDigitalInputThresholdError._build_from_response(
            request, response
        )
    elif error_type == qm.pb.errors_pb2.InvalidDigitalInputWindowError:
        exception_to_raise = InvalidDigitalInputWindowError._build_from_response(
            request, response
        )
    elif error_type == qm.pb.errors_pb2.InvalidDigitalInputPolarityError:
        exception_to_raise = InvalidDigitalInputPolarityError._build_from_response(
            request, response
        )
    else:
        exception_to_raise = UnspecifiedError("Unspecified operation specific error")

    return exception_to_raise



class FailedToExecuteJobException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class FailedToAddJobToQueueException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class CompilationException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class JobCancelledError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class InvalidStreamMetadataError(Exception):
    def __init__(self, *args: object):
        super().__init__(*args)


class ConfigValidationException(Exception):
    def __init__(self, *args: object):
        super().__init__(*args)


class ConfigSerializationException(Exception):
    def __init__(self, *args: object):
        super().__init__(*args)



import json as _json
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from typing import Tuple, Union

import numpy
import numpy.lib.format as _format
from google.protobuf.json_format import MessageToDict

from qm._QmJobErrors import (
    _handle_job_manager_error,
    MissingElementError,
    ElementWithoutIntermediateFrequencyError,
    ElementWithSingleInputError,
    InvalidElementCorrectionError,
    MissingDigitalInputError,
    InvalidDigitalInputThresholdError,
    InvalidDigitalInputWindowError,
    InvalidDigitalInputPolarityError,
)
from qm._report import ExecutionReport
from qm._results import JobResults
from qm.logger import logger
from qm.pb.frontend_pb2 import (
    IsJobRunningRequest,
    IsJobAcquiringDataRequest,
    IsJobAcquiringDataResponse,
    HaltRequest,
    GetSimulatedQuantumStateRequest,
)
from qm.pb.frontend_pb2 import ResumeRequest, PausedStatusRequest
from qm.pb.inc_qua_pb2 import QuaProgram
from qm.pb.job_manager_pb2 import (
    SetElementCorrectionRequest,
    GetElementCorrectionRequest,
    SetDigitalInputThresholdRequest,
    GetDigitalInputThresholdRequest,
    SetDigitalInputWindowRequest,
    GetDigitalInputWindowRequest,
    SetDigitalInputPolarityRequest,
    ACTIVE_HIGH,
    ACTIVE_LOW,
    GetDigitalInputPolarityRequest,
)
from qm.pb.job_manager_pb2_grpc import JobManagerServiceStub
from qm.pb.job_results_pb2 import PullSimulatorSamplesRequest
from qm.pb.job_results_pb2_grpc import JobResultsServiceStub
from qm.persistence import BinaryAsset
from qm.results.SimulatorSamples import SimulatorSamples
from qm.utils import insert_input_stream_to_job


class AcquiringStatus(Enum):
    AcquireStopped = 0
    NoDataToAcquire = 1
    HasDataToAcquire = 2


@dataclass
class DensityMatrix:
    timestamp: int
    data: numpy.ndarray


class SimulatorOutput:
    def __init__(self, job_id: str, frontend) -> None:
        super().__init__()
        self._id = job_id
        self._frontend = frontend

    def get_quantum_state(self) -> DensityMatrix:
        request = GetSimulatedQuantumStateRequest()
        request.jobId = self._id

        result = self._frontend.GetSimulatedQuantumState(request)
        if result.ok:
            flatten = numpy.array(
                [complex(item.re, item.im) for item in result.state.data]
            )
            n = int(numpy.sqrt(len(flatten)))
            if len(flatten) != (n * n):
                raise RuntimeError("Quantum state matrix is not correct")
            else:
                pass

            matrix = flatten.reshape(n, n)
            timestamp = result.state.timeStamp
            return DensityMatrix(timestamp, matrix)
        else:
            raise RuntimeError("Error while pulling quantum state")


class QmJob:
    def __init__(
        self,
        qmm: "qm.QuantumMachinesManager.QuantumMachinesManager",
        job_id: str,
        execute_response=None,
    ):
        super().__init__()
        self._id = job_id
        self._logger = logger.getChild("job")

        self._simulated_analog_outputs = {"samples": None, "waveforms": None}
        self._simulated_digital_outputs = {"samples": None, "waveforms": None}
        if execute_response is not None and execute_response.HasField("simulated"):
            if execute_response.simulated.HasField("analogOutputs"):
                self._simulated_analog_outputs = MessageToDict(
                    execute_response.simulated.analogOutputs
                )
            if execute_response.simulated.HasField("digitalOutputs"):
                self._simulated_digital_outputs = MessageToDict(
                    execute_response.simulated.digitalOutputs
                )
            if len(execute_response.simulated.errors) > 0:
                raise RuntimeError("\n".join(execute_response.simulated.errors))

        self._qmm = qmm
        self._frontend = qmm._frontend
        self._job_manager_client = JobManagerServiceStub(self.manager._channel)

    def id(self):
        """
        :return: The id of the job
        """
        return self._id

    @property
    def manager(self) -> "qm.QuantumMachinesManager.QuantumMachinesManager":
        """
        The QM object where this job lives

        :return:

        """
        return self._qmm

    def simulated_analog_waveforms(self):
        """
        Return the results of the simulation of quantum elements and analog outputs.

        The returned dictionary has the following keys and entries:

        ``elements``: a dictionary containing the outputs with timestamps and values arranged by quantum elements.

        ``controllers``: a dictionary containing the outputs with timestamps and values arranged by controllers.

            ``ports``: a dictionary containing the outputs with timestamps and values arranged by output ports.

                for each element or output port, the entry is a list of dictionaries with the following information:

                ``timestamp``:

                    The time, in nsec, from the start of the program to the start of the pulse.

                ``samples``:

                    Output information, with ``duration`` given in nsec and ``value`` given normalized OPX output units.


        :return: A dictionary containing output information for the analog outputs of the controller.

        """
        return self._simulated_analog_outputs["waveforms"]

    def simulated_digital_waveforms(self):
        """
        Return the results of the simulation of digital outputs.

        ``controllers``: a dictionary containing the outputs with timestamps and values arranged by controllers.

            ``ports``: a dictionary containing the outputs with timestamps and values arranged by output ports.

                for each element or output port, the entry is a list of dictionaries with the following information:

                ``timestamp``:

                    The time, in nsec, from the start of the program to the start of the pulse.

                ``samples``:

                    A list containing the sequence of outputted values, with ``duration`` given in nsec
                    and ``value`` given as a boolean value.



        :return: A dictionary containing output information for the analog outputs of the controller.
        """
        return self._simulated_digital_outputs["waveforms"]

    def halt(self) -> bool:
        """
        Halts the job on the opx
        """
        request = HaltRequest()
        request.jobId = self._id

        response = self._frontend.Halt(request)
        return response.ok

    def _get_np_simulated_samples(
        self, path=None, include_analog=True, include_digital=True
    ):
        request = PullSimulatorSamplesRequest()
        request.jobId = self._id
        request.includeAnalog = include_analog
        request.includeDigital = include_digital

        writer = BytesIO()
        data_writer = BytesIO()

        for result in self._frontend.PullSimulatorSamples(request):
            if result.ok:
                one_of = result.WhichOneof("body")
                if one_of == "header":
                    _format.write_array_header_2_0(
                        writer,
                        {
                            "descr": _json.loads(result.header.simpleDType),
                            "fortran_order": False,
                            "shape": (result.header.countOfItems,),
                        },
                    )
                elif one_of == "data":
                    data_writer.write(result.data.data)

            else:
                raise RuntimeError("Error while pulling samples")

        data_writer.seek(0)
        for d in data_writer:
            writer.write(d)

        writer.seek(0)
        return numpy.load(writer)

    def get_simulated_samples(self, include_analog=True, include_digital=True):
        """
        Obtain the output samples of a QUA program simulation.

        Samples are returned in an object that holds the controllers in the current simulation,
        where each controller's name will be a property of this object.
        The value of each property of the returned value is an object with the following properties:

        ``analog``:

            holds a dictionary with analog port names as keys and numpy array of samples as values.

        ``digital``:

            holds a dictionary with digital port names as keys and numpy array of samples as values.

        It is also possible to directly plot the outputs using a built-in plot command.

        Example::

            >>> samples = job.get_simulated_samples()
            >>> analog1 = samples.con1.analog["1"]  # obtain analog port 1 of controller "con1"
            >>> digital9 = samples.con1.analog["9"]  # obtain digital port 9 of controller "con1"
            >>> samples.con1.plot()  # Plots all active ports
            >>> samples.con1.plot(analog_ports=['1', '2'], digital_ports=['9'])  # Plots the given output ports

        .. note::

            The simulated digital waveforms are delayed by 136ns relative to the real
            output of the OPX.

        :param include_analog: Should we collect simulated analog samples
        :param include_digital: Should we collect simulated digital samples

        :return: The simulated samples of the job.
        """
        return SimulatorSamples.from_np_array(
            self._get_np_simulated_samples(
                include_analog=include_analog, include_digital=include_digital
            )
        )

    def resume(self):
        """
        Resumes a program that was halted using the pause statement
        """
        request = ResumeRequest()
        request.jobId = self._id
        self._frontend.Resume(request)

    @property
    def result_handles(self) -> JobResults:
        """
        :type: qm._results.JobResults
        :return: The handles that this job generated
        """
        has_job_streaming_state = self.manager._caps.has_job_streaming_state
        if not has_job_streaming_state:
            import qm._results_01

            return qm._results_01.JobResults(
                self._id,
                JobResultsServiceStub(self.manager._channel),
                self.manager.store,
            )
        return JobResults(
            self._id, JobResultsServiceStub(self.manager._channel), self.manager.store
        )

    @property
    def simulator(self) -> SimulatorOutput:
        return SimulatorOutput(self._id, self._frontend)

    def set_element_correction(
        self, element, correction
    ) -> Tuple[float, float, float, float]:
        """
        Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a quantum element.

        Changes will only be done to the current job!

        Values will be rounded to an accuracy of 2^-16.
        Valid values for the correction values are between -2 and (2 - 2^-16).

        Warning - the correction matrix can increase the output voltage which might result in an
        overflow.

        :param element: the name of the element to update the correction for
        :type element: str
        :param correction:

            tuple is of the form (v00, v01, v10, v11) where
            the matrix is
            | v00 v01 |
            | v10 v11 |

        :type correction: tuple
        :return: The correction matrix, after rounding to the OPX resolution.
        """
        request = SetElementCorrectionRequest()
        request.jobId = self._id
        request.qeName = element
        request.correction.v00 = correction[0]
        request.correction.v01 = correction[1]
        request.correction.v10 = correction[2]
        request.correction.v11 = correction[3]

        response = self._job_manager_client.SetElementCorrection(request)
        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            InvalidElementCorrectionError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def get_element_correction(self, element) -> Tuple[float, float, float, float]:
        """
        Gets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a quantum element.

        :param element: the name of the element to update the correction for
        :type element: str
        :return: The current correction matrix
        """
        request = GetElementCorrectionRequest()
        request.jobId = self._id
        request.qeName = element

        response = self._job_manager_client.GetElementCorrection(request)
        valid_errors = (
            MissingElementError,
            ElementWithSingleInputError,
            ElementWithoutIntermediateFrequencyError,
        )
        _handle_job_manager_error(request, response, valid_errors)
        return (
            response.correction.v00,
            response.correction.v01,
            response.correction.v10,
            response.correction.v11,
        )

    def is_paused(self):
        """
        :return: Returns ``True`` if the job is in a paused state.

        see also:

            ``resume()``
        """
        request = PausedStatusRequest()
        request.jobId = self._id

        response = self._frontend.PausedStatus(request)
        return response.isPaused

    def _is_job_running(self):
        """
        :return: Returns ``True`` if the job is running
        """
        request = IsJobRunningRequest()
        request.jobId = self._id

        response = self._frontend.IsJobRunning(request)
        return response.isRunning

    def _is_data_acquiring(self):
        """
        Returns the data acquiring status.
        The possible statuses are: AcquireStopped, NoDataToAcquire,  HasDataToAcquire

        :return: An AcquiringStatus enum object
        """
        request = IsJobAcquiringDataRequest()
        request.jobId = self._id

        response = self._frontend.IsJobAcquiringData(request)
        result = response.acquiringStatus
        if result == IsJobAcquiringDataResponse.ACQUIRE_STOPPED:
            return AcquiringStatus(0)
        elif result == IsJobAcquiringDataResponse.NO_DATA_TO_ACQUIRE:
            return AcquiringStatus(1)
        elif result == IsJobAcquiringDataResponse.HAS_DATA_TO_ACQUIRE:
            return AcquiringStatus(2)

    def execution_report(self) -> ExecutionReport:
        """
        Get runtime errors report for this job. See :ref:`Guides/error:Runtime Errors`.

        :type: qm._report.ExecutionReport
        :return: An object holding the errors that this job generated.
        """
        return ExecutionReport(self._id, JobResultsServiceStub(self.manager._channel))

    def set_digital_input_threshold(
        self, port: Tuple[str, int], threshold: float
    ) -> float:
        request = SetDigitalInputThresholdRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]
        request.threshold = threshold

        response = self._job_manager_client.SetDigitalInputThreshold(request)
        valid_errors = (MissingDigitalInputError, InvalidDigitalInputThresholdError)
        _handle_job_manager_error(request, response, valid_errors)
        return response.threshold

    def get_digital_input_threshold(self, port: Tuple[str, int]) -> float:
        request = GetDigitalInputThresholdRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]

        response = self._job_manager_client.GetDigitalInputThreshold(request)
        valid_errors = (MissingDigitalInputError,)
        _handle_job_manager_error(request, response, valid_errors)
        return response.threshold

    def set_digital_input_window(self, port: Tuple[str, int], window: int) -> int:
        request = SetDigitalInputWindowRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]
        request.window = window

        response = self._job_manager_client.SetDigitalInputWindow(request)
        valid_errors = (MissingDigitalInputError, InvalidDigitalInputWindowError)
        _handle_job_manager_error(request, response, valid_errors)
        return response.window

    def get_digital_input_window(self, port: Tuple[str, int]) -> int:
        request = GetDigitalInputWindowRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]

        response = self._job_manager_client.GetDigitalInputWindow(request)
        valid_errors = (MissingDigitalInputError,)
        _handle_job_manager_error(request, response, valid_errors)
        return response.window

    def set_digital_input_polarity(self, port: Tuple[str, int], polarity: str) -> str:
        request = SetDigitalInputPolarityRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]
        request.polarity = ACTIVE_HIGH
        if polarity == "HIGH":
            request.polarity = ACTIVE_HIGH
        elif polarity == "LOW":
            request.polarity = ACTIVE_LOW
        else:
            raise InvalidDigitalInputPolarityError(
                f"Invalid value for polarity {polarity}. Valid values are: 'HIGH' "
                f"or 'LOW'"
            )

        response = self._job_manager_client.SetDigitalInputPolarity(request)
        valid_errors = (MissingDigitalInputError, InvalidDigitalInputPolarityError)
        _handle_job_manager_error(request, response, valid_errors)
        return response.polarity

    def get_digital_input_polarity(self, port: Tuple[str, int]) -> str:
        request = GetDigitalInputPolarityRequest()
        request.jobId = self._id
        request.digitalInputPort.SetInParent()
        request.digitalInputPort.controllerName = port[0]
        request.digitalInputPort.portNumber = port[1]

        response = self._job_manager_client.GetDigitalInputPolarity(request)
        valid_errors = (MissingDigitalInputError,)
        _handle_job_manager_error(request, response, valid_errors)
        if response.polarity == ACTIVE_HIGH:
            return "HIGH"
        elif response.polarity == ACTIVE_LOW:
            return "LOW"
        else:
            raise Exception(
                f"Received invalid digital input polarity value {response.polarity} from server."
            )

    def insert_input_stream(
        self,
        input_stream: Union[
            QuaProgram.ArrayVarRefExpression, QuaProgram.AnyScalarExpression
        ],
        data,
    ):
        """
        Insert data to the running job input stream declared in the QUA program.
        The data is then ready to be read by the program using the advance
        input stream QUA statement.
        Note - multiple data entries can be inserted before the data is read by the
        program.

        :param input_stream: The input stream the data is to be inserted to.
        :param data: The data to be inserted.
        """
        insert_input_stream_to_job(self, data, input_stream)


class _SavedResults:
    def __init__(self, asset: BinaryAsset) -> None:
        self._asset = asset
        super().__init__()

    def get_numpy_results(self, *args, **kwargs):
        input = self._asset.for_reading()
        try:
            return numpy.load(input)
        finally:
            input.close()



import traceback
import os
from pathlib import Path


def _get_loc():
    qm_package_dir = Path(os.path.abspath(__file__)).parent
    trace = [
        i
        for i in traceback.extract_stack()
        if qm_package_dir not in Path(i.filename).parents
    ][-1]
    return 'File "{}", line {}: {} '.format(trace.filename, trace.lineno, trace.line)
